import json
import linecache
import re
import sys
import re
import traceback

# from .render_html import render

from .parse_code import parse_find_name, parse_find_app, parse_find_index, parse_find_infix

DEFINED_ERRORS = []


def KOGI_ERR(**kw):
    global DEFINED_ERRORS
    defined = dict(**kw)
    if isinstance(defined['pattern'], str):
        defined['pattern'] = re.compile(defined['pattern'])
    if 'keys' in defined:
        defined['keys'] = tuple(defined['keys'].split(
            ',')) if ',' in defined['keys'] else (defined['keys'],)
    DEFINED_ERRORS.append(defined)

def DEBUG_ERR(**kw):
    global DEFINED_ERRORS
    defined = dict(**kw)
    if isinstance(defined['pattern'], str):
        defined['pattern'] = re.compile(defined['pattern'])
    if 'keys' in defined:
        defined['keys'] = tuple(defined['keys'].split(
            ',')) if ',' in defined['keys'] else (defined['keys'],)
    DEFINED_ERRORS = [defined] + DEFINED_ERRORS


# KOGI 定義

LANG_NAME = {
    'true': 'True',
    'false': 'False',
    'null': 'None',
    'NULL': 'None',
}

MODULE_ALIAS = {
    'math': 'math',
    'pd': 'pandas',
    'np': 'numpy',
}

MODULE_NAME = {
    'sin': 'math',
    'cos': 'math',
}


def check_name(slots, line):
    name = slots['matched']['name']
    slots['line'] = line
    if name in LANG_NAME:
        pyname = LANG_NAME[name]
        slots['translated'] = f'{name}は、Pythonでは{pyname}と書きます'
        slots['reason'] = f'別の言語の記法を使っている可能性が高いです'
        slots['hint'] = f'Pythonに早く慣れましょう'
        slots['solution'] = f'{name}を{pyname}に置き換えてください'
        return
    if name in MODULE_ALIAS:
        slots['reason'] = 'モジュールがインポートされていない'
        module_name = MODULE_ALIAS[name]
        slots['hint'] = f'モジュール名はたぶん {module_name}'
        if module_name == name:
            slots['solution'] = f'import {module_name}'
        else:
            slots['solution'] = f'import {module_name} as {name}'
        return
    ss = parse_find_name(line, name)
    id = 'モジュール'
    if len(ss) == 1:
        id = '変数もしくは定数' if ss[0]['ctype'] == 'var' else '関数'
    if name in MODULE_NAME:
        slots['reason'] = f'{id}がインポートされていない'
        module_name = MODULE_NAME[name]
        slots['hint'] = f'モジュール名はたぶん {module_name}'
        slots['solution'] = f'from {module_name} import {name}'
        return
    if id.startswith('変数'):
        slots['reason'] = f'{name}の単なる打ち間違い\nもしくは値を一度も代入していない'
        slots['hint'] = f'{name}の値は何ですか？考えてみよう'
        slots['solution'] = f'{name} = ... のように定義する'
    elif id.startswith('関数'):
        slots['reason'] = f'{name}の単なる打ち間違い\nもしくは関数やクラスを定義していない'
        slots['hint'] = f'実行忘れのセルがないか、確認してみよう'
    else:
        slots['hint'] = f'{name}の種類（変数、関数名、クラス名、モジュール名）を確認しましょう',


def test_NameError():
    print(undefined)


def test_NameError2():
    pd.read_csv('')


def test_NULL():
    print(null)


KOGI_ERR(
    pattern='name \'(.*?)\' is not defined',
    keys='name',
    translated='名前エラーです。\n{name}という名前を使いましたが、未定義です。\nつまり使える状態ではありません',
    check=check_name,
    reason='{name}の単なる打ち間違いかも',
    hint='{name}の種類（変数、関数名、モジュール名）を確認しましょう',
    solution='{name}の種類によって解決法は異なります',
    test=(test_NameError, test_NameError2, test_NULL),
)


def check_recv(slots, line):
    name = slots['matched']['name']
    type = slots['matched']['type']
    slots['line'] = line
    ss = parse_find_name(line, name, defined_key='recv')
    if len(ss) == 1:
        recv = ss[0]['recv']
        id = 'メソッド' if ss[0]['ctype'] == 'method' else 'プロパティ'
        slots['translated'] = f'型エラー: {recv}は{type}型です。{name}のような{id}はありません'
        slots['reason'] = f'{name}の打ち間違い、もしくは{recv}の{type}型が想定外'
        slots['hint'] = f'{recv}にが想定通りの型となるように修正してください'


def test_NoAttribute():
    d = {'a': 1}
    print(d.a)


def test_NoMethod():
    d = {'a': 1}
    print(d.a())


KOGI_ERR(
    pattern='\'(.*?)\' object has no attribute \'(.*?)\'',
    keys='type,name',
    translated='型エラー: レシーバは{type}型です。{name}のようなメソッドやプロパティはありません.',
    reason='{name}の打ち間違い、もしくはレシーバの型{type}が想定外',
    hint='レシーバが想定通りの型となるように修正してください',
    check=check_recv,
    test=(test_NoAttribute, test_NoMethod),
)


def test_ModuleNoAttribute():
    import math as m
    m.sins(1)


KOGI_ERR(
    pattern='module \'(.*?)\' has no attribute \'(.*?)\'',
    keys='module,name',
    translated='{module}モジュールには、{name}のような関数やプロパティはありません',
    reason='{name}の打ち間違いか、もしくは{module}を別のモジュールと勘違いしている',
    hint='dir({module})で利用できる名前を調べてみよう.',
    test=test_ModuleNoAttribute,
)


def test_ImportError():
    from math import sins


KOGI_ERR(
    pattern='ImportError: cannot import name \'(.*?)\' from \'(.*?)\'',
    keys='name,module',
    translated='{module}モジュールには、{name}のような関数やプロパティはありません',
    reason='{name}の打ち間違いか、もしくは{module}を別のモジュールと勘違いしている',
    hint='dir({module})で利用できる名前を調べてみよう.',
    test=test_ImportError,
)


def test_UnsupportedOperand():
    print(1+'2')


KOGI_ERR(
    pattern='unsupported operand type\(s\) for (.*?): \'(.*?)\' and \'(.*?)\'',
    keys='name,type,type2',
    translated='{type}と{type2}の間で計算しようとしたけど、そこでは演算子{name}は使えません.',
    solution='{type}か{type2}のどちらかの型に変換するといいかも',
    # inspect=inspect_infix,
    error_ext='{left}と{right}の間で計算しようとしたけど、型がそれぞれ{type}と{type2}で異なり、演算子{name}は使えません.',
    test=test_UnsupportedOperand,
)

KOGI_ERR(
    pattern='\'(.*?)\' not supported between instances of \'(.*?)\' and \'(.*?)\'',
    keys='name,type,type2',
    translated='{type}と{type2}の間で計算しようとしたけど、そこでは演算子{name}は使えません.',
    solution='{type}か{type2}のどちらかの型に変換するといいかも',
    # inspect=inspect_infix,
    error_ext='{left}と{right}の間で計算しようとしたけど、型がそれぞれ{type}と{type2}で異なり、演算子{name}は使えません.',
)


def check_callable(slots, lines):
    import builtins
    type = slots['matched']['type']
    ss = parse_find_app(lines)
    #print(ss)
    if len(ss) == 1 and 'cname' in ss[0]:
        name = ss[0]['cname']
        slots['translated'] = f'{name}の値は{type}型の値が代入されています。\n関数ではありません'
        slots['reason'] = f'{name}を変数名として代入してしまうと、関数として使えなくなります'
        slots['hint'] = f'{name}を何とかして関数に戻します'
        if name in dir(builtins):
            slots['solution'] = f'`from buitins import {name}`を試して'

def test_NotCallable():
    a = 1
    a()


KOGI_ERR(
    pattern='\'(.*?)\' object is not callable',
    keys='type',
    translated='{type}は、関数ではありません. ',
    hint='関数名を変数名として使ってしまうと発生します.',
    solution='環境を初期化してみてください',
    check=check_callable,
    test=test_NotCallable,
)


def test_NotIterable():
    list(1)


KOGI_ERR(
    pattern='\'(.*?)\' object is not iterable',
    keys='type',
    translated='{type}は、イテラブルではありません. つまり、列に変換できませんし、for文などで繰り返し処理もできません.',
    test=test_NotIterable,
)


def test_NotSubscriptable():
    a = 1
    a[0]


KOGI_ERR(
    pattern='\'(.*?)\' object is not subscriptable',
    keys='type',
    translated='{type}は、データ列でもマッピングでもありません.',
    reason='たぶん、x[y]のように操作したけど、xは[]で値を取れません。',
    test=test_NotSubscriptable,
)


def test_MustBeNoneOr():
    print(end=1)


KOGI_ERR(
    pattern='(\\w+?) must be None or a (\\w+?), not (\\w+)',
    keys='name,type,type2',
    translated='{name}は、Noneか{type}の値です. {type2}の値は使えません.',
    test=test_MustBeNoneOr,
)

#render_value() missing 1 required positional argument: 'value'
#TypeError: render_value() missing 2 required positional arguments: 'typename' and 'value'

KOGI_ERR(
    pattern='(\\S+?)\\(\\) missing (\\d+) required positional arguments?: (.*)$',
    keys='name,num,arguments',
    translated='型エラーです\n{name}には、{num}個の引数が足りません',
    reason='足りない引数は、{arguments}です。',
    solution='{name}の定義をよくみてください',
    #check=check_arguments,
)

def test_InvalidKeyword():
    print(color='red')


KOGI_ERR(
    pattern='\'(.*?)\' is an invalid keyword argument for (.*?)\\(\\)',
    keys='name,funcname',
    translated='{name}は、{funcname}の引数として使えるキーワードではありません.',
    solution='{funcname}のリファレンスを読んで確認してください.',
    test=test_InvalidKeyword,
)

KOGI_ERR(
    pattern='(\\S+?)\\(\\) got an unexpected keyword argument \'(.*?)\'',
    keys='funcname,name',
    translated='{name}は、{funcname}の引数として使えるキーワードではありません.',
    solution='{funcname}のリファレンスを読んで確認してください.',
)


KOGI_ERR(
    pattern='invalid (syntax)',
    keys='dummy',
    translated='構文エラーです',
    reason='Pythonは構文規則通りに書かなければなりません',
    hint='エラーの発生した前の行に問題があるかも',
)

KOGI_ERR(
    pattern='expected (an) indented block',
    keys='dummy',
    translated='構文エラー: インデントが足りません',
    reason='空白とタブが混ざっているかも',
    hint='エラーの発生した前の行に問題があるかも',
)

KOGI_ERR(
    pattern='unexpected (indent)',
    keys='dummy',
    translated='構文エラー: インデントが余分です',
    reason='空白とタブが混ざっているかも',
    hint='エラーの発生した前の行に問題があるかも',
)

KOGI_ERR(
    pattern='unindent does (not) match any outer indentation level',
    keys='dummy',
    translated='構文エラー: インデントの深さが変で、どのブロックに属すのかわかりません',
    reason='空白とタブが混ざっているかも',
    hint='エラーの発生した前の行に問題があるかも',
    solution='インデントの深さを揃えます',
)

KOGI_ERR(
    pattern='invalid (character) in identifier',
    keys='dummy',
    translated='構文エラー: 使用できない文字が使われています',
    reason='全角文字が混ざっている可能性が高いです',
    hint='先生やTAさんに質問する案件ではありません.',
    solution='落ち着いて打ち直そう',
)

KOGI_ERR(
    pattern='unexpected (EOF) while parsing',
    keys='dummy',
    translated='構文エラー: コードが途中までしか書かれていません. ',
    reason='たぶん、括弧やクオートを閉じ忘れている可能性が高いです.',
    hint='友達にチェックしてもらうとエラーが見つかるかも',
    solution='エラーの発生した行の前後を含めて、構文エラーを探して！',
)


def test_InvalidLiteral():
    int('a')


KOGI_ERR(
    pattern='invalid literal for int\(\) with base (.*?): (\'.*?\')',
    keys='base,value',
    translated='文字列{value}は、整数に変換できる文字列ではありません',
    test=test_InvalidLiteral,
)


def test_NotConvert():
    float('a')


KOGI_ERR(
    pattern='could not convert (\w+?) to (\w+?): \'(.*?)\'',
    keys='type,type2,value',
    translated='{type}の{value}は、{type2}に変換することはできません.',
    test=test_NotConvert,
)


def test_OutOfRange():
    s = "ABC"
    s[3]


KOGI_ERR(
    pattern='(\\w+) index out of range',
    keys='type',
    translated='インデックスが{type}の大きさを超えています. ',
    # nspect=inspect_index,
    error_ext='インデックス{index}が{callee}の大きさを超えています. ',
    solution_ext='インデックス{index}がlen({callee})未満に収まるように条件を加えてください. ',
    test=test_OutOfRange,
)


def test_KeyError():
    s = {}
    s["A"]


KOGI_ERR(
    pattern='KeyError: (.+?)$',
    keys='key',
    translated='キー{key}が見つかりません',
    # inspect=inspect_index,
    error_ext='{callee}にはキー{key}がありません. ',
    reason_ext='{index}が未定義なキー{key}として与えられています. ',
    solution_ext='{callee}のキーを確認しましょう. ',
    test=test_KeyError,
)


def test_MustBeInteger():
    s = "ABC"
    s['3']


KOGI_ERR(
    pattern='(\\w+) indices must be integers',
    keys='type',
    translated='{type}のインデックスは、整数でなければなりません.',
    # inspect=inspect_index,
    error_ext='{callee}のインデックスは、整数でなければなりません. ',
    reason_ext='{index}が整数でありません. ',
    solution_ext='{index}を整数に変換してみては？ ',
    test=test_MustBeInteger,
)


def test_DividedByZero():
    print(1/0)


KOGI_ERR(
    pattern='division by (zero)',
    keys='dummy',
    translated='ゼロで徐算しました. ',
    reason='ゼロで割ることはできません. ',
    hint='分母の値を確認してみましょう.',
    test=test_DividedByZero
)


def test_FileNotFoundError():
    open('file')


KOGI_ERR(
    pattern='FileNotFoundError: \\[Errno (\\d+)\\] No such file or directory: (\'.*\')',
    keys='errno,file',
    translated='{file}が見つかりません. ',
    reason='ファイル名やファイルパスが間違っている. ',
    hint='{file}が存在するか確認しよう.',
    test=test_FileNotFoundError,
)

# UsageError: Cell magic `%%kogi2` not found.

KOGI_ERR(
    pattern='UsageError: Cell magic `(.*)` not found',
    keys='name',
    translated='セルマジック{name}は、まだ使えません. ',
    reason='セルマチックをインポートし忘れています。',
    hint='何をインポートすべきか、{name}をググって探しましょう.',
)

# UsageError: %%kogi is a cell magic, but the cell body is empty.

KOGI_ERR(
    pattern='UsageError: (\\S.) is a cell magic, but the cell body is empty',
    keys='name',
    translated='{name}はセルマジックなので、本体が必要です',
    reason='セルマジックに続くべき、コードが空です',
    solution='セルの本文を書きましょう',
)


# ValueError: too many values to unpack (expected 2)

KOGI_ERR(
    pattern='ValueError: too many values to unpack \\(expected (\\d.)\\)',
    keys='num',
    translated='値エラー: {num}個の値を変数に展開しようとしたが、個数が不一致です。',
    reason='リストやタプルなどの値の数が、想定通りではない',
    hint='入力データも確認してみてください',
)


# error


def _format(ss, vocab):
    if isinstance(ss, str):
        return ss.format(**vocab)
    return [s.format(**vocab) for s in ss]


def _copy_and_format(defined, key, slots):
    if key in defined:
        slots[key] = _format(defined[key], slots['matched'])


def parse_error_message(code, emsg, lines):
    if emsg.startswith('Kogi'):
        _, _, data = emsg.partition(':')
        return json.loads(data)
    for defined in DEFINED_ERRORS:
        matched_result = defined['pattern'].search(emsg)
        if matched_result:
            matched = {}
            slots = dict(emsg=emsg, keys=defined['keys'], matched=matched)
            try:
                for i, key in enumerate(defined['keys']):
                    matched[key] = matched_result.group(i+1)
            finally:
                _copy_and_format(defined, 'translated', slots)
                _copy_and_format(defined, 'reason', slots)
                _copy_and_format(defined, 'hint', slots)
                _copy_and_format(defined, 'solution', slots)
            if 'check' in defined:
                try:
                    defined['check'](slots, lines)
                except:
                    traceback.print_exc()
            slots['code'] = code
            return slots
    return {'code': code, 'emsg': emsg}


def exception_stack(etype, evalue, tb):
    stacks = []
    while tb:
        filename = tb.tb_frame.f_code.co_filename
        if 'lib' not in filename:
            name = tb.tb_frame.f_code.co_name
            lineno = tb.tb_lineno
            line = linecache.getline(filename, lineno, tb.tb_frame.f_globals)
            local_vars = tb.tb_frame.f_locals
            stacks.append(dict(filename=filename,
                               etype=f'{etype.__name__}',
                               emsg=f'{etype.__name__}: {evalue}',
                               name=name, lineno=lineno,
                               line=line, vars=local_vars))
        tb = tb.tb_next
    return list(stacks[::-1])


def parse_exception(code=None, print=print):
    etype, evalue, tb = sys.exc_info()
    emsg = f'{etype.__name__}: {evalue}'
    stacks = exception_stack(etype, evalue, tb)
    lines = [stack['line'].strip() for stack in stacks]
    return parse_error_message(code, emsg, lines)


def test_exception_hook():
    slots = parse_exception()
    print(slots)


if __name__ == '__main__':
    def test_DEFINED_ERRORS():
        for defined in DEFINED_ERRORS:
            fs = defined.get('test', ())
            if not isinstance(fs, tuple):
                fs = (fs,)
            for f in fs:
                try:
                    f()
                except:
                    test_exception_hook()

    test_DEFINED_ERRORS()



import os
from re import M

import sys
from .logger import kogi_print
from .nmt_compose import compose

DEVICE = None
model = None
tokenizer = None
cached = {}


def _load_gdown(model_path, model_id, quiet=True):
    kogi_print('Downloading Kogi Programming AI Model...')
    os.system('pip install --upgrade gdown')
    import gdown
    url = f'https://drive.google.com/uc?id={model_id}'

    gdown.download(url, 'model.zip', quiet=quiet)
    os.system(f'rm -rf {model_path}')
    os.system(f'unzip -d {model_path} -j model.zip')


def load_model(model_id, model_path='./kogi_model'):
    global model, tokenizer, cached, DEVICE
    if not os.path.exists(model_path):
        _load_gdown(model_path=model_path, model_id=model_id)

    if os.path.exists(model_path):
        kogi_print('Initializing Transormers and T5 ...')
        try:
            import sentencepiece
        except ModuleNotFoundError:
            os.system('pip install sentencepiece')
        import torch
        try:
            from transformers import T5ForConditionalGeneration, MT5Tokenizer
        except ModuleNotFoundError:
            os.system('pip install transformers')
            from transformers import T5ForConditionalGeneration, MT5Tokenizer

        USE_GPU = torch.cuda.is_available()
        DEVICE = torch.device('cuda:0' if USE_GPU else 'cpu')
        kogi_print('DEVICE :', DEVICE)

        model = T5ForConditionalGeneration.from_pretrained(model_path)
        tokenizer = MT5Tokenizer.from_pretrained(model_path, is_fast=True)
        cached = {}


def greedy_search(s: str, max_length=128, beam=1) -> str:
    input_ids = tokenizer.encode_plus(
        s,
        add_special_tokens=True,
        max_length=max_length,
        padding="do_not_pad",
        truncation=True,
        return_tensors='pt').input_ids.to(DEVICE)
    greedy_output = model.generate(input_ids, max_length=max_length)
    return tokenizer.decode(greedy_output[0], skip_special_tokens=True)


# def beam_search(s, max_length=128, beams=5):
#     input_ids = tokenizer.encode_plus(
#         s,
#         add_special_tokens=True,
#         max_length=max_length,
#         padding="do_not_pad",
#         truncation=True,
#         return_tensors='pt').input_ids.to(DEVICE)
#     beam_outputs = model.generate(
#         input_ids,
#         max_length=max_length,
#         num_beams=beams,
#         no_repeat_ngram_size=2,
#         num_return_sequences=beams,
#         early_stopping=True
#     )
#     return [tokenizer.decode(beam_output, skip_special_tokens=True) for beam_output in beam_outputs]


# def _translate_beam(s: str, beams: int, max_length=64):
#     global model, tokenizer, DEVICE
#     model.config.update({"num_beams": beams})
#     input_ids = tokenizer.encode_plus(s,
#                                       add_special_tokens=True,
#                                       max_length=max_length,
#                                       # padding="max_length",
#                                       padding="do_not_pad",
#                                       truncation=True,
#                                       return_tensors='pt').input_ids.to(DEVICE)
#     predict = model.generate(input_ids,
#                              return_dict_in_generate=True,
#                              output_scores=True,
#                              length_penalty=8.0,
#                              max_length=max_length,
#                              num_return_sequences=beams,
#                              early_stopping=True)
#     pred_list = sorted([[tokenizer.decode(predict.sequences[i], skip_special_tokens=True),
#                          predict.sequences_scores[i].item()] for i in range(len(predict))], key=lambda x: x[1], reverse=True)
#     sentences_list = [i[0] for i in pred_list]
#     scores_list = [i[1] for i in pred_list]
#     return sentences_list, scores_list


model_id = None


def kogi_enable_ai(access_key: str, start_loading=False):
    global model_id
    model_id = access_key
    if model_id is not None:
        load_model(model_id)


def get_nmt():
    global model, cached
    if model_id is None:
        return compose(lambda s: 'わん')
    if model is None:
        load_model(model_id)

    def generate(s, max_length=80):
        if s in cached:
            return cached[s]
        t = greedy_search(s, max_length=max_length)
        cached[s] = t
        return t
    return compose(generate)


if __name__ == '__main__':
    nmt = get_nmt()
    for a in sys.argv[1:]:
        print(a, nmt(a))




# try:
#     from .webui import kogi_login, kogi_help, kogi_translate, kogi_print
# except ModuleNotFoundError:
#     pass
#from .runner import kogi, corgi


try:
    from .exception_hook import enable_kogi_hook, disable_kogi_hook
    from .parse_error import DEBUG_ERR
    enable_kogi_hook()
except ModuleNotFoundError:
    pass



import uuid
import json
from datetime import datetime
import requests

verbose = True

def kogi_verbose(enabled: bool):
    global verbose
    verbose = enabled


def kogi_print(*args, **kw):
    global verbose
    if verbose:
        print('\033[35m[🐶]', *args, **kw)
        print('\033[0m', end='')


def print_nop(*x, **kw):
    pass

## LOGGER

slack = None

def load_slack(slack_id='QNZDoUPuLo7C3lHyh9PWhZz3', update=True):
    global slack
    if slack_id is None:
        slack = None
        return None
    try:
        from slackweb import Slack     
    except ModuleNotFoundError:
        import os
        os.system('pip install slackweb')
        from slackweb import Slack     
    HOST = 'slack.com'
    ID = 'T02NYCBFP7B'
    ID2 = 'B02QPM8HNBH'
    if '/' not in slack_id:
        url = f'https://hooks.{HOST}/services/{ID}/{ID2}/{slack_id}'
    else:
        url = f'https://hooks.{HOST}/services/{slack_id}'
    try:
        local_slack =  Slack(url)
        if update:
            slack = local_slack
    except Exception as e:
        print('Slack Error', e)
    return local_slack
    

def send_slack(logs, slack_id='QNZDoUPuLo7C3lHyh9PWhZz3'):
    try:
        local_slack = load_slack(slack_id=slack_id, update=False)
        jsondata = json.dumps(logs, ensure_ascii=False)
        local_slack.notify(text = jsondata)
    except Exception as e:
        print('Slack Error:', e)



SESSION = str(uuid.uuid1())
SEQ = 0
LOGS = []
UID = 'unknown'
KEY='OjwoF3m0l20OFidHsRea3ptuQRfQL10ahbEtLa'
epoch = datetime.now().timestamp()


def send_log(right_now=False, print=kogi_print):
    global epoch, LOGS
    url = 'https://ixe8peqfii.execute-api.ap-northeast-1.amazonaws.com/dev'
    now = datetime.now().timestamp()
    delta = (now - epoch)
    epoch = now
    if len(LOGS) > 0 and (right_now or delta > 180):
        data = {
            "session": SESSION,
            "uid": UID,
            "logs": LOGS.copy(),
        }
        LOGS.clear()
        headers = {'x-api-key': f'A{KEY}s'}
        r = requests.post(url, headers=headers, json=data)
        if r.status_code != 200:
            print(data)

def log(**kw):
    global SEQ, LOGS, epoch
    now = datetime.now()
    date = now.isoformat(timespec='seconds')
    logdata = dict(seq=SEQ, date=date, **kw)
    LOGS.append(logdata)
    SEQ += 1
    send_log()
    return logdata

def record_login(uid, **kw):
    global UID
    UID = f'{uid}'
    logdata = log(uid=UID, **kw)
    send_log(right_now=True)
    if slack is None:
        send_slack(logdata)
        

if __name__ == '__main__':
    # log(a=1, b=2)
    # log_now()
    record_login(uid='11111', test='test')



from .utils import zen2han, listfy
from .render_html import render, render_value, render_astype
from .dialog_desc import response_desc
from .nmt import get_nmt

REMOVED = [
    '.', '。', '?', '？', '！',
    '何', '何ですか', '何でしょうか',
    'が知りたい', 'がしりたい', 'がわからない', 'が分からない',
]


def remove_tails(s):
    for suffix in REMOVED:
        if s.endswith(suffix):
            return remove_tails(s[:-len(suffix)])
    return s


def get_global_frame():
    return {}


class Chatbot(object):
    frame: dict
    render_html: bool

    def __init__(self, frame=None, render_html=True):
        self.frame = {} if frame is None else frame
        self.frame.update(get_global_frame())
        self.render_html = render_html
        self.nmt = get_nmt()

    def response(self, text):
        text = zen2han(text)
        text = remove_tails(text)
        if text.endswith('には'):
            text = text[:-2]
            return self.response_translate(text)
        if text.endswith('って') or text.endswith('とは'):
            text = text[:-2]
            return self.response_desc(text)
        if text.startswith('原因') or text.startswith('理由') or text.startswith('なぜ') or text.startswith('なんで'):
            if 'reason' in self.frame:
                return self.frame['reason']
            else:
                return self.response_vow(text)
        if text.startswith('解決') or text.startswith('どう'):
            if 'solution' in self.frame:
                return self.frame['solution']
            else:
                if 'reason' in self.frame:
                    return '原因を特定してみてね'
                return 'ググってみたら'
        if text.startswith('ヒント'):
            if 'hint' in self.frame:
                return self.frame['hint']
            else:
                return 'ノー ヒント！'
        return self.response_code(text)

    def response_translate(self, text):
        return self.nmt(f'trans: {text}')

    def response_desc(self, text):
        return response_desc(text)

    def response_code(self, text):
        try:
            v = get_ipython().ev(text)
            self.frame['code'] = text
            self.frame['value'] = v
        except:
            pass
        if 'value' not in self.frame:
            return self.response_vow(text)
        code = render(text, 'code', render_html=self.render_html)
        tyname = render_astype(v, render_html=self.render_html)
        value = render_value(v, render_html=self.render_html)
        return [f'{code}の型は{tyname}。値は', value]

    def response_vow(self, text):
        return self.nmt(f'talk: {text}')


def get_chatbot(frame=None):
    chatbot = Chatbot(frame=frame)
    return lambda text: chatbot.response(text)


if __name__ == '__main__':
    chat = get_chatbot()
    print(chat('イテラブルとは'))



import os
from .logger import log

DESC = None


def load_descriptions():
    global DESC
    if DESC is None:
        # base = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')
        # for file in os.listdir(base):
        #     if file.endswith(.csv)
        # pass
        DESC = {
        "コギー": "倉光研究室のAI犬", 
		"プログラミング": "コンピュータに意図した動作を行わせるために処理手順を作成すること",
		"コーディング": "プログラミング言語に従って、コンピュータが処理・解釈できるコードを書くこと",
		"仕様書": "ソフトウェアやサービスなどを開発する前に、機能やその性能、満たすべき要件など記述した文書のこと",
		"アルゴリズム": "ある特定の問題を解く手順を計算や操作の組み合わせとして明確に書いたもの",
		"機械語": "コンピュータのマイクロプロセッサ（CPU）が直接解釈・実行できる命令コードのこと",
		"オペコード": "マイクロプロセッサ（CPU/MPU）に与える機械語の命令の識別番号のこと",
		"オペランド": "数式を構成する要素のうち、演算の対象となる値や変数のこと",
		"演算子": "数学やプログラミングなどで式を記述する際に用いられる、演算内容を表す記号などのこと",
		"単項演算子": "一つの項を用いる演算子のこと",
		"二項演算子": "二つの項を用いる演算子のこと",
		"三項演算子": "条件式の真偽に応じて二つの値のいずれかを返す条件演算子のこと",
		"算術演算子": "四則演算を行う演算子のこと",
		"論理演算子": "論理演算を表す記号や符号などのこと。",
		"XOR": "排他的論理和のこと",
		"排他的論理和": "二つの命題のいずれか一方のみが真のときに真となり、両方真や両方偽のときは偽となる論理演算",
		"論理否定": "与えられた命題が真のときに偽となり、偽のとき真となる論理演算",
		"論理和": "二つの命題のいずれか一方あるいは両方が真のときに真となり、いずれも偽のときに偽となる論理演算",
		"論理積": "二つの命題のいずれも真のときに真となり、それ以外のときは偽となる論理演算",
		"真理値": "ある命題が真（True）であるか、偽(False）であるかを示す値のこと。",
		"真偽値": "ある命題が真（True）であるか、偽(False）であるかを示す値のこと。",
		"ブール値": "ある命題が真（True）であるか、偽(False）であるかを示す値のこと。",
		"ビット演算": "整数を2進数のビット列とみなして、ビットの移動やビット単位の論理演算などを行うもの。",
		"インクリメント": "数値を増やす操作のこと",
		"デクリメント": "数値を減らす操作のこと",
		"定数": "ある特定の数値やデータに名前を与えたもの。変更はできない。",
		"変数": "データを一時的に記憶しておくための領域に固有の名前を付けたもの。",
		"スコープ": "プログラム中で変数名を参照できる範囲のこと。",
		"グローバル変数": "プログラム中のどこのコードからでも参照できる変数のこと。バグの元と言われる。",
		"ローカル変数": "宣言された関数やメソッドなどの内部からしか参照できない変数のこと。",
		"関数": "数学の関数を模したサブルーチンで引数に対し処理し、結果を返すもの。",
		"サブルーチン": "特定の機能や処理をひとまとまりにした小プログラムで、他のプログラムから呼び出して実行できるもの",
		"戻り値": "関数やメソッドが処理を終えるときに、呼び出し元に対して渡すデータのこと。return文から返される。",
		"返り値": "関数やメソッドが処理を終えるときに、呼び出し元に対して渡すデータのこと。return文から返される。",
		"ビルトイン関数": "あらかじめ用意された関数のこと。",
		"ユーザ定義関数": "ユーザが定義した関数のこと。",
		"標準ライブラリ関数": "ビルトイン関数ではないが、標準的に提供されるユーザ定義関数のこと。",
		"入出力": "入力と出力のこと。I/Oとも。",
		"浮動小数点数": "コンピュータにおける数値の表現形式の一つで、小数点以下の値を含む数値の表現法として最も広く利用されている。",
		"単精度浮動小数点数": "32ビット長の浮動小数点数を格納できる数値のこと。floatとも。",
		"倍精度浮動小数点数": "64ビット長の浮動小数点数を格納できる数値のこと。doubleとも。",
		"文字列": "文字を並べたデータのこと。",
		"データ": "何かを文字や符号、数値などのまとまりとして表現したもの。",
		"データ構造": "データの集まりをプログラムで扱いやすいようにまとめたもの。",
		"タプル": "順序付けられた複数の要素で構成されるデータ構造のこと。数学に由来。",
		"配列": "複数のデータを連続的に並べたデータ構造のこと。効率が良い。",
		"多次元配列": "配列の要素に配列が入った入れ子状の配列のこと。",
		"二次元配列": "",
		"動的配列": "配列の長さが固定的されておらず、実行時に要素を追加、削除できる配列のこと。",
		"可変長配列": "配列の長さが固定的されておらず、実行時に要素を追加、削除できる配列のこと。",
		"添字": "配列に格納された個々の要素の位置を示す値のこと。数列の添字に由来。インデックスとも。",
		"インデックス": "配列に格納された個々の要素の位置を示す値のこと。数列の添字に由来。",
		"イテレータ": "配列のようなデータ構造の要素を順に走査していく繰り返し処理を簡潔に記述できるオブジェクトのこと。",
		"連想配列": "Pythonの辞書に相当するデータ構造のこと",
		"リスト": "複数のデータを順序を付けて格納するデータ構造のこと",
		"プレフィックス": "文字列や番号の先頭に付加して特定の意味を付け加える要素のこと。接頭辞とも。",
		"サフィックス": "文字列や番号の先頭に付加して特定の意味を付け加える要素のこと。接尾辞とも。",
		"基数": "位取り記数法で数値を書き記す際に各桁の重み付けの基本となる数で、位が上がる毎に何倍になるかを表す。",
		"リテラル": "ソースコード上で、特定のデータ型の値を直に記載したもの",
		"即値": "ソースコード中に直に書き込まれたデータのこと",
		"書式": "文書や書類の定められた様式やテンプレート",
		"フォーマット": "形式、書式、様式、体裁、型、構成などの意味を持つ",
		"テンプレート": "鋳型、雛型、定型書式などの意味を持つ英単語。",
		"インデント": "字下げのこと。文の行頭に空白を挿入して読みやすくすること。Pythonのようにインデントによってプログラムの構造を記述する。",
		"マジックナンバー": "ソースコードに直書きされた数値。その意味や意図が記述した本人以外わからなくなる恐れあり。",
		"NaN": "数値として表せない数値。非数とも。",
		"オブジェクト": "互いに密接に関連するデータと手続き（処理手順）をひとまとまりにしたもののこと。",
		"クラス": "オブジェクト指向プログラミングにおけるオブジェクトの雛型のこと。",
		"インスタンス": "あらかじめ定義されたクラスなどをメインメモリ上に展開して処理・実行できる状態にした実体のこと。",
		"フィールド": "オブジェクトの持つ固有のデータのこと。メンバ変数やプロパティとも",
		"メソッド": "各オブジェクトに属する処理や操作のこと。",
		"カプセル化": "外部に対して必要な情報のみ提供し、外から直に参照や操作する必要のないプロパティやメソッドを秘匿すること。",
		"ペイロード": "送受信されるデータのパケットのうち、宛先などの制御情報を除いたデータ本体のこと",
		"パディング": "データの長さを一定にするため前後に挿入される余白や無意味なデータのこと。",
		"ゼロパディング": "数値を文字として表示する際に、固定帳の桁数に合わせて左右に「0」を追加すること。",
		"固定長": "データや要素、領域などの長さがあらかじめ決まって変化しないこと。対義語は「可変長」",
		"可変長": "データや要素、領域などの長さが実行時に変化させられること。",
		"引数": "関数やメソッドなどを呼び出すときに渡すデータのこと。",
		"パラメータ": "ソフトウェアやシステムの挙動に影響を与える外部から与えられるデータのこと。",
		"仮引数": "関数やメソッドが呼び出し元から渡された値を受け取るために宣言された変数のこと。",
		"実引数": "関数やメソッドを呼び出す際に実際に引き渡される値や変数のこと。",
		"可変長引数": "あらかじめ数が固定されてなく、任意個の引数を受け取れること",
		"ステップ数": "コンピュータプログラムの規模を測る指標の一つ。実質処理を行っているソースコード行数のこと。",
		"API": "あるサービスを外部のプログラムから呼び出して利用できる規約のこと。",
		"デバッグ": "コンピュータプログラムに潜む欠陥を探し出して取り除くこと。",
		"バグ": "はプログラムに含まれる誤りのこと。必ずある。",
		"バグる": "意図しない奇妙な挙動を示すこと。",
		"コンパイル": "ソースコードをコンピュータが実行可能な形式に変換すること。",
		"コンパイラ": "ソースコードをコンピュータが実行可能な形式に変換するツールのこと。",
		"スクリプト言語": "コンパイルなしで実行できる軽量なプログラミング言語のこと。",
		"プラットフォーム": "あるソフトウェアを動作させるのに必要な土台となる装置やソフトウェア、サービスのこと。",
		"クロスプラットフォーム": "あるソフトウェアを複数の異なる仕様の機種やOSで同じように動作させられること。",
		"VM": "ソフトウェアで実現されたコンピュータやマイクロプロセッサの動作を模した仮想的なコンピュータのこと。",
		"仮想マシン": "ソフトウェアで実現されたコンピュータやマイクロプロセッサの動作を模した仮想的なコンピュータのこと。",
		"インタプリタ": "ソースコードをコンピュータが解釈・実行できる形式に変換しながら実行するソフトウェアのこと。",
		"動的": "プログラムの実行中に状態や構成が決まること。対義語は「静的」",
		"静的": "プログラムの実行前に状態や構成が決まること。対義語は「動的」",
		"オーバーヘッド": "目的の処理を行うために発生する余分な処理やコストのこと。",
		"動作環境": "ソフトウェアが動作するハードウェアやOSの構成や設定のこと",
		"環境": "ソフトウェアが動作するハードウェアやOSの構成や設定のこと",
		"ビルド": "ソースコードなどを元に実行可能ファイルや配布パッケージを作成すること",
		"デプロイ": "開発したソフトウェアを実際の運用環境に配置・展開して実用に提供すること",
		"ポーティング": "あるシステムで動作するよう開発されたソフトウェアを、異なる仕様や設計のシステムで動作するように作り変えること。",
		"アルゴリズム": "ある特定の問題を解く手順を、単純な計算や操作の組み合わせとして明確に定義したもの。",
		"エラー": "プログラムの実行を継続できない致命的な支障のこと。（頑張って取り除こう！）",
		"ランタイムエラー": "コンピュータプログラムの実行時に発生するエラーのこと。",
		"コンパイル": "プログラミング言語で書かれたコンピュータプログラム（ソースコード）を解析し、コンピュータが直接実行可能な形式のプログラム（オブジェクトコード）に変換すること。",
		"ライブラリ": "ある機能を持ったプログラムを利用できるように部品化し、集めたプログラムのこと。",
		"ファイル": "ストレージなどにデータを記録する際の最小の記録単位となるデータのまとまり。",
		"ファイル名": "コンピュータのストレージ上に保存されたファイルの名前",
		"拡張子": "ファイル名のうち、「.」で区切られた右側の部分でファイルの種類を表す",
		"ファイルシステム": "ストレージの記録状態を管理・制御し、ファイル単位でデータ操作可能にするOSの機能のこと",
		"ディレクトリ": "複数のファイルをまとめて分類できる名前の付いた保管場所のこと。フォルダとも。",
		"フォルダ": "複数のファイルをまとめて分類できる名前の付いた保管場所のこと。ディレクトリとも。",
		"カレントディレクトリ": "実行中のソフトウェアが現在ファイル操作をできるディレクトリのこと",
		"パス": "コンピュータシステム内で特定のリソースの位置を表す文字列のこと",
		"ファイルパス": "ストレージ内のファイルやディレクトリの位置を表す文字列のこと",
		"相対パス": "現在位置からの相対的な位置によってパスを記述する記法のこと。",
		"絶対パス": "階層構造のルートからの位置によってパスを記述する記法のこと。",
		"ストレージ": "永続的にデータを保存できる記憶装置のこと。ハードディスクやSSD、USBメモリなど。",
		"シェル": "利用者からのコマンドを受け付けて結果を提示を行うOSの機能のこと",
		"CLI": "コマンドベースのユーザーインターフェースのこと",
		"プロンプト": "シェルにおいて入力を受け付けられる状態であることを示すメッセージのこと",
		"マクロ": "ソースコード中に繰り返し登場する特定の記述を、別の短い記述に置き換える機能のこと",
		"プリプロセッサ": "ある中心的な処理を行うプログラムに対して、その前処理を行うプログラムのこと。",
		"文字コード": "文字や記号を扱うために割り当てられた番号のこと。",
		"Unicode": "世界中の様々な言語の文字を同時に使えるように規格化された文字コードのこと。ユニコードと読む。",
		"ユニコード": "世界中の様々な言語の文字を同時に使えるように規格化された文字コードのこと。",
		"UTF-8": "UnicodeをASCIIコードと互換のある形式で表現できる文字コードのこと。",
		"互換のある形式": "同時に扱っても不具合が生じない形式のこと",
		"バイト": "8ビットのことを表す情報量の単位のこと",
		"バイト列": "任意のビットパターンからなるバイト単位の長さを持つデータの集合。",
		"エンディアン": "多バイトデータを記録するときの順序のこと。ビックエンディアンとリトルエンディアンがある。",
		"ビッグエンディアン": "多バイトデータを最上位から下位への順序で扱う方式。",
		"リトルエンディアン": "多バイトデータを最下位から上位への順序で扱う方式。",
		"スラッシュ": "右上から左下への斜め線「/」文字のこと。",
		"バックスラッシュ": "左上から右下への斜め線「\」のこと。スラッシュの逆向き。円マークが代わりに使われることもある。",
		"全角文字": "等幅フォントで幅と高さが等しい文字のこと。要するに、漢字やひらがなのこと。",
		"半角文字": "全角文字の半分の幅に表示される文字のこと。要するに、英数字などのASCII文字のこと。",
		"ASCII文字": "ASCII(American Standard Code for Information Interchange)規格の文字。アスキーと読む。",
		"制御文字": "通信制御などの制御に用いるように割り当てられた文字コード。タブや改行が該当する。",
		"エスケープ文字": "制御文字など印字しにくい文字コードを数値などで表記する記法のこと。",
		"エスケープシーケンス": "制御文字など印字しにくい文字コードを数値などで表記する記法のこと。",
		"EOF": "ファイルの終端のこと。",
    }
    return DESC


def response_desc(text):
    desc = load_descriptions()
    if text in desc:
        return text
    log(type='undefined_desc', query=text)
    return 'わん'



try:
    import pegtree as pg
    from pegtree.visitor import ParseTreeVisitor
except ModuleNotFoundError:
    import os
    os.system('pip3 install pegtree')
    import pegtree as pg
    from pegtree.visitor import ParseTreeVisitor


def fix(tree):
    a = [tree.epos_]
    for t in tree:
        a.append(fix(t).epos_)
    for key in tree.keys():
        a.append(fix(tree.get(key)).epos_)
    #print(repr(tree), a)
    tree.epos_ = max(a)
    return tree


def fix2(s):
    if not isinstance(s, str):
        s = str(fix(s))
    open = s.count('(')
    close = s.count(')')
    if open > 0 and open > close:
        return s+(')' * (open-close))
    open = s.count('[')
    close = s.count(']')
    if open > 0 and open > close:
        return s+(']' * (open-close))
    return s


parser = pg.generate(pg.grammar('inspect.pegtree'))


class Flatten(ParseTreeVisitor):
    def __init__(self):
        ParseTreeVisitor.__init__(self)

    def parse(self, s: str, bufs=None):
        tree = parser(s)
        self.bufs = [] if bufs is None else bufs
        self.visit(tree)
        return self.bufs

    def acceptName(self, tree):
        self.bufs.append(dict(
            cname=str(tree),
            ctype='var',
        ))

    def acceptInfix(self, tree):
        self.bufs.append(dict(
            cname=str(tree.name).strip(),
            ctype='infix',
            infix_left=str(fix2(tree.left)),
            infix_right=str(fix2(tree.right)),
        ))
        self.visit(tree.left)
        self.visit(tree.right)

    def acceptApplyExpr(self, tree):
        self.bufs.append(dict(
            cname=fix2(tree.name),
            ctype='app',
        ))
        self.visit(tree.params)

    def acceptGetExpr(self, tree):
        self.bufs.append(dict(
            cname=str(tree.name),
            ctype='field',
            recv=fix2(tree.recv),
        ))

    def acceptMethodExpr(self, tree):
        self.bufs.append(dict(
            cname=str(tree.name),
            ctype='method',
            recv=fix2(tree.recv),
        ))
        self.visit(tree.params)

    def acceptIndexExpr(self, tree):
        self.bufs.append(dict(
            ctype='index',
            recv=fix2(tree.recv),
            index=fix2(tree.index),
        ))
        self.visit(tree.recv)
        self.visit(tree.index)

    def acceptUndefined(self, tree):
        for t in tree:
            self.visit(t)
        for key in tree.keys():
            self.visit(tree.get(key))


f = Flatten()


def parse_find(line, ctype=None, cname=None, defined_key=None, ctype_key='ctype', cname_key='cname'):
    rs = f.parse(line.strip())
    ss = []
    for r in rs:
        if defined_key is not None and defined_key not in r:
            continue
        if cname is not None and r.get(cname_key, None) != cname:
            continue
        if ctype is not None and r.get(ctype_key, None) != ctype:
            continue
        r['code'] = line.strip()
        ss.append(r)
    return ss


def parse_find_name(lines, name, defined_key='cname'):
    ss = []
    for line in lines:
        ss.extend(parse_find(line, defined_key=defined_key, cname=name))
    return ss

def parse_find_app(lines):
    ss = []
    for line in lines:
        ss.extend(parse_find(line, ctype='app'))
    return ss

def parse_find_infix(lines, infix):
    ss = []
    for line in lines:
        ss.extend(parse_find(line, ctype='infix', cname=infix))
    return ss


def parse_find_index(lines):
    ss = []
    for line in lines:
        ss.extend(parse_find(line, ctype='index'))
    return ss


if __name__ == '__main__':
    print(parse_find_name('a+b', 'a'))
    print(parse_find_name('a()+b', 'a'))
    print(parse_find_name('o.a+b', 'a'))
    print(parse_find_name('o.a()+b', 'a'))
    print(parse_find_name('a.f+b', 'a'))
    print(parse_find_name('a[1]+b', 'a'))
    print(parse_find_name('s[a]+b', 'a'))
    print('infix')
    print(parse_find_infix('a+b+c', '+'))
    print(parse_find_infix('a()+b', '+'))
    print(parse_find_infix('o.a+b', '+'))
    print(parse_find_infix('o.a()+b', '+'))
    print(parse_find_infix('a.f+b', '+'))
    print(parse_find_infix('a[b+1]+b', '+'))
    print(parse_find_infix('s[a]+b', '+'))
    print('index')
    print(parse_find_index('a[b+1]+b'))
    print(parse_find_index('s[a]+b'))
    print(parse_find_index('a[b+1:]+b'))
    print(parse_find_index('s[a:x]+b'))



BOT_ICON = 'https://iconbu.com/wp-content/uploads/2021/02/コーギーのイラスト.jpg'
#BOT_ICON = 'https://kohacu.com/wp-content/uploads/2021/05/kohacu.com_samune_003370-768x768.png'
YOUR_ICON = 'https://2.bp.blogspot.com/-VVtgu8RyEJo/VZ-QWqgI_wI/AAAAAAAAvKY/N-xnZvqeGYY/s800/girl_question.png'

CHAT_CSS = '''
<style>
.sb-box { position: relative; overflow: hidden; }
/* アイコン画像 */
.icon-img {
    position: absolute; overflow: hidden; top: 0; width: 64px; height: 64px;
}
/* アイコン画像（左） */
.icon-img-left { left: 0; }
/* アイコン画像（右） */
.icon-img-right { right: 0; }
/* アイコン画像 */
.icon-img img { border-radius: 50%; border: 2px solid #eee; }
/* アイコンネーム */
.icon-name {
    position: absolute; width: 64px; text-align: center;
    top: 68px; color: #fff; font-size: 10px;
}
/* アイコンネーム（左） */
.icon-name-left { left: 0; }
/* アイコンネーム（右） */
.icon-name-right { right: 0; }
/* 吹き出し */
.sb-side { position: relative; float: left; margin: 0 85px 20px 85px; }
.sb-side-right { float: right; }
/* 吹き出し内のテキスト */
.sb-txt {
    position: relative; border: 2px solid #eee; border-radius: 6px;
    background: #eee; color: #333;
    font-size: 15px; line-height: 1.7; padding: 18px;
}
.sb-txt>p:last-of-type { padding-bottom: 0; margin-bottom: 0; }
/* 吹き出しの三角 */
.sb-txt:before { content: ""; position: absolute; border-style: solid; top: 16px; z-index: 3; }
.sb-txt:after { content: ""; position: absolute; border-style: solid; top: 15px; z-index: 2; }
/* 吹き出しの三角（左） */
.sb-txt-left:before {
    left: -7px; border-width: 7px 10px 7px 0;
    border-color: transparent #eee transparent transparent;
}
.sb-txt-left:after {
    left: -10px; border-width: 8px 10px 8px 0;
    border-color: transparent #eee transparent transparent;
}
/* 吹き出しの三角（右） */
.sb-txt-right:before {
    right: -7px; border-width: 7px 0 7px 10px;
    border-color: transparent transparent transparent #eee;
}
.sb-txt-right:after {
    right: -10px; border-width: 8px 0 8px 10px;
    border-color: transparent transparent transparent #eee;
}
.box{ background: powderblue; }
.scrolly{ overflow-y: scroll; }
textarea {
  width: 100%; 
  box-sizing: border-box;  /* ※これがないと横にはみ出る */
  height:30px; 
  font-size: large;
  outline: none; /* ※ブラウザが標準で付加する線を消したいとき */
  resize: none;
}
</style>
'''

CHAT_HTML = '''
<div id='main'>
<script>
var timer = null;
var inputPane = document.getElementById('input');
inputPane.addEventListener('keydown', (e) => {
  if(e.keyCode == 13) {
    var text = inputPane.value;
    google.colab.kernel.invokeFunction('notebook.ask', [text], {});
    inputPane.value='';
    if(timer !== null) {
        clearTimeout(timer);
    }
    timer = setTimeout(()=>{
        google.colab.kernel.invokeFunction('notebook.log', [], {});
    }, 1000*60*5);
  }
});
var target = document.getElementById('output');
target.scrollIntoView(false);
</script>
<div id='output' class='box scrolly'>
</div>
<div style='text-align: right'>
<textarea id='input' placeholder='質問はここに'></textarea>
</div>
</div>
'''

BOT_HTML = '''
<div class="sb-box">
    <div class="icon-img icon-img-left">
        <img src="{}" width="60px">
    </div>
    <div class="icon-name icon-name-left">{}</div>
    <div class="sb-side sb-side-left">
        <div class="sb-txt sb-txt-left">
          {}
        </div>
    </div>
</div>
'''

CLEAR_HTML = '''
<script>
setTimeout(()=>{
    const element = document.getElementById('main'); 
    element.remove();
}, 2000);
</script>
'''


USER_HTML = '''
<div class="sb-box">
  <div class="icon-img icon-img-right">
      <img src="{}" width="60px">
  </div>
  <div class="icon-name icon-name-right">{}</div>
  <div class="sb-side sb-side-right">
      <div class="sb-txt sb-txt-right">
        {}
      </div>
  </div>
</div>
'''



from IPython.core.magic import register_cell_magic

try:
    from .webui import kogi_say
except ModuleNotFoundError:
    pass

from .utils import print_nop
from .logger import log, kogi_print
from .dialog import get_chatbot
from .errors import kogi_check_error


KOGI_OPTIONS = [
]


def kogi_add_option(isOption, runner):
    KOGI_OPTIONS.append((isOption, runner))


def _run_cell(code, option):
    res = get_ipython().run_cell(code)
    res.raise_error()


def _kogi_check_option(option):
    for isOption, runner in KOGI_OPTIONS:
        if isOption(option):
            return runner
    return _run_cell


def kogi_run(code, option, run_cell=None):
    if run_cell is None:
        run_cell = _kogi_check_option(option)
    try:
        run_cell(code, option)
    except:
        results = kogi_check_error(code, show=print_nop, render_html=True)
        if 'error' in results:
            log(type='error', error=results)
            kogi_say(results['error'], get_chatbot(results))
        else:
            log(type='undefined_error', error=results)


@register_cell_magic
def kogi(option, code):
    kogi_run(code, option)


@register_cell_magic
def corgi(option, code):
    kogi_run(code, option)



import IPython
from IPython.display import display, HTML
from google.colab import output
from .utils import listfy
from .logger import load_slack, kogi_print, log, send_log, print_nop, record_login
from .nmt import get_nmt, kogi_enable_ai
from .dialog import get_chatbot


def debug_log():
    try:
        send_log()
    except Exception as e:
        kogi_print(e)

# https://github.com/googlecolab/colabtools/tree/0162530b8c7f76741ee3e518db34aa5c173e8ebe/google/colab


BOT_ICON = 'https://iconbu.com/wp-content/uploads/2021/02/コーギーのイラスト.jpg'
#BOT_ICON = 'https://kohacu.com/wp-content/uploads/2021/05/kohacu.com_samune_003370-768x768.png'
YOUR_ICON = 'https://2.bp.blogspot.com/-VVtgu8RyEJo/VZ-QWqgI_wI/AAAAAAAAvKY/N-xnZvqeGYY/s800/girl_question.png'

CHAT_CSS = '''
<style>
.sb-box { position: relative; overflow: hidden; }
/* アイコン画像 */
.icon-img {
    position: absolute; overflow: hidden; top: 0; width: 64px; height: 64px;
}
/* アイコン画像（左） */
.icon-img-left { left: 0; }
/* アイコン画像（右） */
.icon-img-right { right: 0; }
/* アイコン画像 */
.icon-img img { border-radius: 50%; border: 2px solid #eee; }
/* アイコンネーム */
.icon-name {
    position: absolute; width: 64px; text-align: center;
    top: 68px; color: #fff; font-size: 10px;
}
/* アイコンネーム（左） */
.icon-name-left { left: 0; }
/* アイコンネーム（右） */
.icon-name-right { right: 0; }
/* 吹き出し */
.sb-side { position: relative; float: left; margin: 0 85px 20px 85px; }
.sb-side-right { float: right; }
/* 吹き出し内のテキスト */
.sb-txt {
    position: relative; border: 2px solid #eee; border-radius: 6px;
    background: #eee; color: #333;
    font-size: 15px; line-height: 1.7; padding: 18px;
}
.sb-txt>p:last-of-type { padding-bottom: 0; margin-bottom: 0; }
/* 吹き出しの三角 */
.sb-txt:before { content: ""; position: absolute; border-style: solid; top: 16px; z-index: 3; }
.sb-txt:after { content: ""; position: absolute; border-style: solid; top: 15px; z-index: 2; }
/* 吹き出しの三角（左） */
.sb-txt-left:before {
    left: -7px; border-width: 7px 10px 7px 0;
    border-color: transparent #eee transparent transparent;
}
.sb-txt-left:after {
    left: -10px; border-width: 8px 10px 8px 0;
    border-color: transparent #eee transparent transparent;
}
/* 吹き出しの三角（右） */
.sb-txt-right:before {
    right: -7px; border-width: 7px 0 7px 10px;
    border-color: transparent transparent transparent #eee;
}
.sb-txt-right:after {
    right: -10px; border-width: 8px 0 8px 10px;
    border-color: transparent transparent transparent #eee;
}
.box{ background: powderblue; }
.scrolly{ overflow-y: scroll; }
textarea {
  width: 100%; 
  box-sizing: border-box;  /* ※これがないと横にはみ出る */
  height:30px; 
  font-size: large;
  outline: none; /* ※ブラウザが標準で付加する線を消したいとき */
  resize: none;
}
</style>
'''

CHAT_HTML = '''
<div id='main'>
<script>
var timer = null;
var inputPane = document.getElementById('input');
inputPane.addEventListener('keydown', (e) => {
  if(e.keyCode == 13) {
    var text = inputPane.value;
    google.colab.kernel.invokeFunction('notebook.ask', [text], {});
    inputPane.value='';
    if(timer !== null) {
        clearTimeout(timer);
    }
    timer = setTimeout(()=>{
        google.colab.kernel.invokeFunction('notebook.log', [], {});
    }, 1000*60*5);
  }
});
var target = document.getElementById('output');
target.scrollIntoView(false);
</script>
<div id='output' class='box scrolly'>
</div>
<div style='text-align: right'>
<textarea id='input' placeholder='質問はここに'></textarea>
</div>
</div>
'''

BOT_HTML = '''
<div class="sb-box">
    <div class="icon-img icon-img-left">
        <img src="{}" width="60px">
    </div>
    <div class="icon-name icon-name-left">{}</div>
    <div class="sb-side sb-side-left">
        <div class="sb-txt sb-txt-left">
          {}
        </div>
    </div>
</div>
'''

CLEAR_HTML = '''
<script>
setTimeout(()=>{
    const element = document.getElementById('main'); 
    element.remove();
}, 8000);
</script>
'''


USER_HTML = '''
<div class="sb-box">
  <div class="icon-img icon-img-right">
      <img src="{}" width="60px">
  </div>
  <div class="icon-name icon-name-right">{}</div>
  <div class="sb-side sb-side-right">
      <div class="sb-txt sb-txt-right">
        {}
      </div>
  </div>
</div>
'''


def _display_bot(bot_text, **kw):
    with output.redirect_to_element('#output'):
        bot_name = kw.get('bot_name', 'コギー')
        bot_icon = kw.get('bot_icon', BOT_ICON)
        for text in listfy(bot_text):
            display(HTML(BOT_HTML.format(bot_icon, bot_name, text)))
    if 'バイバイ' in bot_text:
        display(HTML(CLEAR_HTML))


def _display_you(your_text, **kw):
    with output.redirect_to_element('#output'):
        your_name = kw.get('your_name', 'あなた')
        your_icon = kw.get('your_icon', YOUR_ICON)
        for text in listfy(your_text):
            display(HTML(USER_HTML.format(your_icon, your_name, text)))


kogi_frame = {  # グローバルフレーム
    'your_name': 'あなた',
    'your_icon': YOUR_ICON,
    'bot_name': 'コギー',
    'bot_icon': BOT_ICON,
    'display': _display_bot,
}


N_GLOBALS = 0


def _needs_new_chat():
    global N_GLOBALS
    state = get_ipython().ev('len(globals())')
    if state != N_GLOBALS:
        N_GLOBALS = state
        return True
    return False


def _display_chat(chatbot=None):
    display(HTML(CHAT_CSS))
    display(HTML(CHAT_HTML))
    if chatbot is None:
        chatbot = get_chatbot()

    def ask(your_text):
        global kogi_frame
        your_text = your_text.strip()
        if 'ありがとう' in your_text or 'バイバイ' in your_text:
            _display_bot('バイバイ')
        else:
            bot_text = chatbot(your_text)
            log(type='chat', user=your_text, bot=bot_text)
            _display_you(your_text, **kogi_frame)
            if bot_text is not None:
                _display_bot(bot_text, **kogi_frame)

    output.register_callback('notebook.ask', ask)
    output.register_callback('notebook.log', debug_log)


def kogi_say(msg, chatbot=None):
    if _needs_new_chat():
        _display_chat(chatbot)
    _display_bot(msg, **kogi_frame)


def kogi_help(chatbot=None):
    kogi_say('どうした？', chatbot)


# translate
TRANSLATE_CSS_HTML = '''
<style>
.parent {
  background-color: #edebeb;
  width: 100%;
  height: 150px;
}
textarea {
  width: 100%; 
  box-sizing: border-box;  /* ※これがないと横にはみ出る */
  height:120px; 
  font-size: large;
  outline: none;           /* ※ブラウザが標準で付加する線を消したいとき */
  resize: none;
}
.box11{
//    padding: 0.5em 1em;
//    margin: 2em 0;
    color: #5d627b;
    background: white;
    border-top: solid 5px #5d627b;
    box-shadow: 0 3px 5px rgba(0, 0, 0, 0.22);
}
.box18{
  //padding: 0.2em 0.5em;
  //margin: 2em 0;
  color: #565656;
  background: #ffeaea;
  background-image: url(https://2.bp.blogspot.com/-u7NQvQSgyAY/Ur1HXta5W7I/AAAAAAAAcfE/omW7_szrzao/s800/dog_corgi.png);
  background-size: 150%;
  background-repeat: no-repeat;
  background-position: top right;
  background-color:rgba(255,255,255,0.8);
  background-blend-mode:lighten;
  //box-shadow: 0px 0px 0px 10px #ffeaea;
  border: dashed 2px #ffc3c3;
  //border-radius: 8px;
}
.box16{
    //padding: 0.5em 1em;
    //margin: 2em 0;
    background: -webkit-repeating-linear-gradient(-45deg, #f0f8ff, #f0f8ff 3px,#e9f4ff 3px, #e9f4ff 7px);
    background: repeating-linear-gradient(-45deg, #f0f8ff, #f0f8ff 3px,#e9f4ff 3px, #e9f4ff 7px);
}
.box24 {
    position: relative;
    padding: 0.5em 0.7em;
    margin: 2em 0;
    background: #6f4b3e;
    color: white;
    font-weight: bold;
}
.box24:after {
    position: absolute;
    content: '';
    top: 100%;
    left: 30px;
    border: 15px solid transparent;
    border-top: 15px solid #6f4b3e;
    width: 0;
    height: 0;
}
</style>
<div class="parent">
<div style="float: left; width: 48%; text-align: right;">
<label class="box24" for="input">日本語</label>
<textarea id="input" class="box16"></textarea>
</div>
<div style="float: left; width: 48%; text-align: right;">
<label class="box24" for="outout">Python</label>
<textarea id="output" class="box18 python" readonly></textarea>
</div>
</div>
'''

TRANSLATE_SCRIPT = '''
<script>
    var timer = null;
    var logtimer = null;
    document.getElementById('input').addEventListener('input', (e) => {
        var text = e.srcElement.value;
        if(timer !== null) {
            clearTimeout(timer);
        }
        if(logtimer !== null) {
            clearTimeout(logtimer);
        }
        timer = setTimeout(() => {
            timer = null;
            (async function() {
                const result = await google.colab.kernel.invokeFunction('notebook.Convert', [text], {});
                const data = result.data['application/json'];
                const textarea = document.getElementById('output');
                textarea.textContent = data.result;
            })();
        }, 600);  // 何も打たななかったら600ms秒後に送信
        logtimer = setTimeout(() => {
            // logtimer = null;
            google.colab.kernel.invokeFunction('notebook.Logger', [], {});
        }, 60*1000*5); // 5分に１回まとめて送信
    });
</script>
'''


cached = {}


def kogi_translate(delay=600, always_policy=False, print=print_nop):
    nmt = get_nmt()

    def convert(text):
        try:
            ss = []
            for line in text.split('\n'):
                if line not in cached:
                    translated = nmt(
                        line, always_policy=always_policy, print=print)
                    print(line, '=>', translated)
                    cached[line] = translated
                    log(
                        type='realtime-nmt',
                        input=line, output=translated,
                    )
                else:
                    translated = cached[line]
                ss.append(translated)
            text = '\n'.join(ss)
            return IPython.display.JSON({'result': text})
        except Exception as e:
            print(e)
        return e
    output.register_callback('notebook.Convert', convert)
    output.register_callback('notebook.Logger', debug_log)
    display(IPython.display.HTML(TRANSLATE_CSS_HTML))
    SCRIPT = TRANSLATE_SCRIPT.replace('600', str(delay))
    display(IPython.display.HTML(SCRIPT))


# LOGIN


# ダミー関数
LOGIN_HTML = '''
<style>
.parent {
  background-color: #edebeb;
  width: 100%;
  height: 150px;
}
textarea {
  width: 100%; 
  box-sizing: border-box;  /* ※これがないと横にはみ出る */
  height:120px; 
  font-size: large;
  outline: none;           /* ※ブラウザが標準で付加する線を消したいとき */
  resize: none;
}
.box16{
    background: -webkit-repeating-linear-gradient(-45deg, #f0f8ff, #f0f8ff 3px,#e9f4ff 3px, #e9f4ff 7px);
    background: repeating-linear-gradient(-45deg, #f0f8ff, #f0f8ff 3px,#e9f4ff 3px, #e9f4ff 7px);
}
.box18{
  color: #565656;
  background: #ffeaea;
  border: dashed 2px #ffc3c3;
}
.box24 {
    position: relative;
    padding: 0.5em 0.7em;
    margin: 2em 0;
    background: #6f4b3e;
    color: white;
    font-weight: bold;
}
.button02 {
  width: 300px;
  color: #333;
  font-weight: 700;
  background-color: #cccccc;
  border-radius: 50vh;
}
</style>
<label>Student ID</label><input id="name"/>
<span class="button02" id="ok">Ready</span>
<div class="parent">
<div style="float: left; width: 48%; text-align: right;">
<label class="box24" for="outout">Code</label>
<textarea id="output" class="box18 python" readonly>print(math.sin(math.pi/2))
print(["oranges", "tables"])
print(weight / (height * height))
print(x if x >= y else y)
print(s[0].upper() for s in "abcdefg")
</textarea>
</div>
<div style="float: left; width: 48%; text-align: right;">
<label class="box24" for="input">Type In</label>
<textarea id="input" class="box16"></textarea>
</div>
</div>
'''

LOGIN_SCRIPT = '''
<script>
    var timer = null;
    var buffers = [];
    var dict = {};
    const idPane = document.getElementById('name');
    const inputPane = document.getElementById('input');
    var submitted = false;
    var buttonClick = () => {
        var name = idPane.value;
        var value = inputPane.value;
        var text = buffers.join(' ');
        google.colab.kernel.invokeFunction('notebook.login', [name, value, dict, text, window.navigator.userAgent], {});
        (async function() {
            const result = await google.colab.kernel.invokeFunction('notebook.login', [name, value, dict, text, window.navigator.userAgent], {});
            const data = result.data['application/json'];
            document.getElementById('ok').innerText = `出席 平均: ${data.time}ms, 正確さ: ${data.acc}`;
        })();
    };
    var before = new Date().getTime();
    idPane.addEventListener('keydown', (e) => {
        before = new Date().getTime();
        if(idPane.value.length >= 7) {
            document.getElementById('ok').innerText = 'Go';
            return;
        }
    });
    inputPane.addEventListener('keydown', (e) => {
      var now = new Date().getTime();
      if(e.key === ' ') {
        buffers.push(`${now - before} SPACE`);
      }
      else {
        buffers.push(`${now - before} ${e.key}`);
      }
      before = now;
      if(idPane.value.length < 7) {
          inputPane.value = '';
          return;
      }
      dict[e.key] = (dict[e.key] || 0) + 1;
      var size = inputPane.value.length;
      if(size > 10 && dict[')'] >= 8 && dict['i'] >= 10 && dict['t'] >= 10) {
        if(!submitted) {
            submitted = true;
            document.getElementById('ok').innerText = '出席';
            setTimeout(buttonClick, 3000);
        }
      }
      else{
        document.getElementById('ok').innerText = `${size}`;
      }
    });
</script>
'''


def _time(keys):
    times = [int(t) for t in keys.split() if t.isdigit()]
    return (sum(times) - max(times)) // (len(times) - 1)


CODE = '''print(math.sin(math.pi/2))
print(["oranges", "tables"])
print(weight / (height * height))
print(x if x >= y else y)
print(s[0].upper() for s in "abcdefg")'''


def _accuracy(code):
    import difflib
    return difflib.SequenceMatcher(None, code.strip(), CODE).ratio()


def kogi_login(ai_key=None, class_name='unknown', slack_key=None, print=print_nop):
    def login(name, code, counts, keys, useragent):
        try:
            code = code.strip()
            acc = round(_accuracy(code), 3)
            time = round(_time(keys), 3)
            keys = keys.split('\n')[-1]
            print(keys)
            record_login(type='typing',
                         uid=name, class_name=class_name,
                         code=code, keys=keys,
                         mean_time=time, accuracy=acc,
                         counts=counts, browser=useragent)
            return IPython.display.JSON({'acc': acc, 'time': time})
        except Exception as e:
            kogi_print(e)

    output.register_callback('notebook.login', login)
    display(IPython.display.HTML(LOGIN_HTML))
    display(IPython.display.HTML(LOGIN_SCRIPT))
    load_slack(slack_key)
    if ai_key is not None:
        try:
            kogi_enable_ai(ai_key, start_loading=True)
        except Exception as e:
            print('Disabled AI', e)
            kogi_enable_ai(None, start_loading=True)



import sys
from IPython.core.magic import register_cell_magic

from .errors import kogi_translate_error
from .webui import kogi_chat

# _lines = None
# _outputs = None


def _run_fn(code, option):
    res = get_ipython().run_cell(code)
    res.raise_error()


def kogi_run(code, option, run_fn=_run_fn):
    # global _lines, _outputs
    try:
        run_fn(code, option)
    except:
        kogi_translate_error(code, render_html=True)
        kogi_chat()
    # finally:
    #     _lines = None
    #     _outputs = None


kogi_OPTIONS = [
]


def kogi_add_option(isOption, runner):
    '''
    Example:
    isOption: lambda option: 'atcoder.com' in option
    return _run_fn
    '''
    kogi_OPTIONS.append(isOption, runner)


def _kogi_check_option(option):
    for isOption, runner in kogi_OPTIONS:
        if isOption(option):
            return runner
    return _run_fn


@register_cell_magic
def kogi(option, code):
    kogi_run(code, option, run_fn=_kogi_check_option(option))



from .parse_error import parse_error_message
from .utils import listfy, zen2han, remove_suffixes

REMOVED_SUFFIXES = [
    '.', '。', '?', '？', '！',
    '何', '何ですか', '何でしょうか',
    'が知りたい', 'がしりたい', 'がわからない', 'が分からない',
]

def startswith(text, prefixes):
    for prefix in prefixes:
        if text.startswith(prefix):
            return True
    return False

class Chatbot(object):
    slots: dict

    def __init__(self, slots=None):
        self.slots = {} if slots is None else slots

    def get(self, key, value=''):
        return self.slots.get(key, value)

    def response(self, text):
        text = zen2han(text)
        text = remove_suffixes(text, REMOVED_SUFFIXES)
        if text.startswith('ダンプ'):
            return f'{self.slots}'
        if startswith(text, ('変数', '助けて', 'たすけて')):
            return self.response_variables()
        if text.endswith('には'):
            text = text[:-2]
            return self.response_translate(text)
        if text.endswith('って') or text.endswith('とは'):
            text = text[:-2]
            return self.response_desc(text)
        if startswith(text, ('原因', '理由', 'なぜ', 'なんで', 'どうして')):
            if 'reason' in self.slots:
                return self.slots['reason']
            else:
                return self.response_vow(text)
        if startswith(text, ('解決', 'どう')):
            if 'solution' in self.slots:
                return self.slots['solution']
            elif 'hint' in self.slots:
                return self.slots['hint']
            elif 'reason' in self.slots:
                return self.slots['hint']
            return self.response_vow(text)
        if text.startswith('ヒント'):
            if 'hint' in self.slots:
                return self.slots['hint']
            elif 'solution' in self.slots:
                return self.slots['solution']
            else:
                return 'ノー ヒント！'
        return self.response_code(text)

    def response_vow(self, text):
        return "わん"
        # return self.nmt(f'talk: {text}')

    def response_translate(self, text):
        # return self.nmt(f'trans: {text}')
        return self.response_vow(text)

    def response_desc(self, text):
        return response_desc(text)

    def response_variables(self, name=None):
        ss = ['変数を全部、表示するよ']
        for stack in self.get('stacks'):
            if 'vars' not in stack:
                continue
            vars = stack['vars']
            if name is None:
                for n in vars.keys():
                    if n.startswith('_') or n in SKIP_IDS:
                        continue
                    v = vars[n]
                    ty = type(v).__name__
                    if ty in ('module', 'function'):
                        continue
                    ss.append(render_value(n, ty, vars[n]))
            elif name in vars:
                v = vars[name]
                ss.append(render_value(name, type(v).__name__, v))
        return ss

    def response_code(self, text):
        # try:
        #     v = get_ipython().ev(text)
        #     self.slots['code'] = text
        #     self.slots['value'] = v
        # except:
        #     pass
        # if 'value' not in self.slots:
        #     return self.response_vow(text)
        # code = render(text, 'code', render_html=self.render_html)
        # tyname = render_astype(v, render_html=self.render_html)
        # value = render_value(v, render_html=self.render_html)
        # return [f'{code}の型は{tyname}。値は', value]
        return self.response_vow(text)


SKIP_IDS = set([
    'In', 'Out', 'get_ipython', 'exit', 'quit'
])

def render_value(name, typename, value):
    head = f'<b>{name}: {typename}型</b>'
    if hasattr(value, '__len__'):
        v = len(value)
        head += f' <tt>len({name})={v}</tt>'
    body = f'<pre>{repr(value)}</pre>'
    if hasattr(value, '_repr_html_'):
        body = value._repr_html_()
    return f'{head}<br/>{body}'


def get_chatbot_webui():
    import traceback
    from IPython.display import display, HTML
    from google.colab import output
    from .html import BOT_ICON, BOT_HTML, CLEAR_HTML, YOUR_ICON, USER_HTML, CHAT_CSS, CHAT_HTML

    def _display_bot(bot_text, chatbot):
        with output.redirect_to_element('#output'):
            bot_name = chatbot.get('bot_name', 'コギー')
            bot_icon = chatbot.get('bot_icon', BOT_ICON)
            for text in listfy(bot_text):
                text = text.replace('\n', '<br/>')
                display(HTML(BOT_HTML.format(bot_icon, bot_name, text)))
        if 'バイバイ' in bot_text:
            display(HTML(CLEAR_HTML))

    def _display_you(your_text, chatbot):
        with output.redirect_to_element('#output'):
            your_name = chatbot.get('your_name', 'あなた')
            your_icon = chatbot.get('your_icon', YOUR_ICON)
            for text in listfy(your_text):
                text = text.replace('\n', '<br/>')
                display(HTML(USER_HTML.format(your_icon, your_name, text)))

    # def debug_log():
    #     try:
    #         send_log()
    #     except Exception as e:
    #         kogi_print(e)

    def _display_chat(chatbot):
        display(HTML(CHAT_CSS))
        display(HTML(CHAT_HTML))

        def ask(your_text):
            your_text = your_text.strip()
            if 'ありがとう' in your_text or 'バイバイ' in your_text:
                display(HTML(CLEAR_HTML))
            try:
                _display_you(your_text, chatbot)
                bot_text = chatbot.response(your_text)
                if bot_text is not None:
                    _display_bot(bot_text, chatbot)
            except Exception as e:
                _display_bot('コギー内部で深刻なエラーが発生しました。\nエラーレポートを頂けると助かります', chatbot)
                traceback.print_exc()
                
        output.register_callback('notebook.ask', ask)
        #output.register_callback('notebook.log', debug_log)

    def kogi_say(msg, chatbot=None):
        if chatbot is None:
            chatbot = Chatbot()
        _display_chat(chatbot)
        _display_bot(msg, chatbot)

    return kogi_say

##


kogi_say = get_chatbot_webui()


def exception_dialog(code, emsg, stacks):
    lines = [stack['line'].strip() for stack in stacks]
    slots = parse_error_message(code, emsg, lines)
    slots['stacks'] = stacks
    chatbot = Chatbot(slots=slots)
    if 'translated' in slots:
        kogi_say(slots['translated'], chatbot)
    else:
        kogi_say('く〜ん', chatbot)



import re
import traceback
import sys
import re

from .render_html import render

from .error_inspect import inspect_callee, inspect_infix, inspect_funcname, inspect_index

DEFINED_ERRORS = []


def KOGI_ERR(d):
    global DEFINED_ERRORS
    DEFINED_ERRORS.append(d)


def KOGI_ERR2(**kw):
    global DEFINED_ERRORS
    DEFINED_ERRORS.append(dict(**kw))


def _format(ss, results):
    vocab = results
    if isinstance(ss, str):
        return ss.format(**vocab)
    return [s.format(**vocab) for s in ss]


def _formatting(defined, key, ext, results):
    key_ext = key+ext
    if key_ext in defined:
        results[key] = _format(defined[key_ext], results)
        return
    if ext == '' and key in defined:
        results[key] = _format(defined[key_ext], results)
        return


def _check_error(errtype, errmsg, code=None, errlines=None, render_html=True):
    s = f'{errtype}: {errmsg}'
    results = {'error_type': errtype, 'error_orig': s}
    for defined in DEFINED_ERRORS:
        if isinstance(defined['pattern'], str):
            defined['pattern'] = re.compile(defined['pattern'])
            defined['keys'] = tuple(defined['keys'].split(','))
        matched = defined['pattern'].search(s)
        if matched:
            for i, key in enumerate(defined['keys']):
                if key == '':
                    break
                results[key] = render(matched.group(i+1), key, render_html=render_html)
            _ext = ''
            if 'inspect' in defined:
                _ext = defined['inspect'](code, errlines, results)
            if 'error_type' in defined:
                results['error_type'] = defined['error_type']
            _formatting(defined, 'error', _ext, results)
            _formatting(defined, 'reason', _ext, results)
            _formatting(defined, 'solution', _ext, results)
            _formatting(defined, 'hint', _ext, results)
            return results
    return results


LinePat = re.compile(r'line (\d+)')
# print(LinePat.match('aa line 18, in <module>'))


def _get_error_lines():
    ss = []
    formatted_lines = traceback.format_exc().splitlines()
    for i, line in enumerate(formatted_lines):
        # print(i, line)
        # if 'ipython-input' in line and ', in ' in line:
        if ', in ' in line:
            matched = LinePat.search(line)
            if matched:
                # ss.append((int(matched.group(1)), formatted_lines[i+1]))
                ss.append(formatted_lines[i+1])
    return ss[::-1]


def _show_verbose(results):
    if 'error_orig' in results:
        print(results['error_orig'])
        print(' =>', results.get('error'), '')
        if 'reason' in results:
            print(' reason: ', results.get('reason'), '')
        if 'solution' in results:
            print(' solution:', results.get('solution'), '')


def kogi_check_error(code=None, show=_show_verbose, render_html=False):
    exc_type, exc_value, exc_tb = sys.exc_info()
    tb = traceback.format_exception(exc_type, exc_value, exc_tb)
    error_lines = _get_error_lines()
    results = _check_error(
        f'{exc_type.__name__}', exc_value, code, error_lines, render_html=render_html)
    show(results)
    if code is not None:
        results['code'] = code
    results['traceback'] = tb
    return results


# kogi 定義

def test_NameError():
    try:
        print(undefined)
    except:
        kogi_check_error()


KOGI_ERR2(
    pattern='name \'(.*?)\' is not defined',
    keys='name',
    error='{name}という名前を使おうとしましたが、まだ一度も使われていません',
    reason=[
        '{name}の単なる打ち間違い',
        '{name}が変数なら、まだ一度も値を代入していない `{name} = ...`',
        '{name}が関数名やクラス名なら未定義、もしくは定義したセルを実行していない',
        '{name}がモジュールなら正しくインポートされていない `import {name}` ',
    ],
    solution='{name}の種類をちゃんと確認しましょう',
    hint='「{name}をインポートするには？」と聞いてみる',
    test=test_NameError,
)


def test_NoAttribute():
    try:
        d = {'a': 1}
        print(d.a)
    except:
        kogi_check_error()


KOGI_ERR2(
    pattern='\'(.*?)\' object has no attribute \'(.*?)\'',
    keys='type,name',
    error='{type}には、{name}のようなメソッドやプロパティはありません.',
    reason='{name}の打ち間違い、もしくは間違った値が使われている.',
    solution='間違った値が代入された箇所を探してください.',
    inspect=inspect_callee,
    error_ext='{callee}は、{type}です. {name}のような{pyname}はありません.',
    reason_ext='{name}{pyname}の打ち間違い、もしくは{callee}に間違った型の値が代入されている.',
    solution_ext='{callee}に間違った値を代入した箇所を探してください.',
    test=test_NoAttribute,
)


def test_ModuleNoAttribute():
    try:
        import math as m
        m.sins(1)
    except:
        kogi_check_error()


KOGI_ERR2(
    pattern='module \'(.*?)\' has no attribute \'(.*?)\'',
    keys='module,name',
    error='{module}モジュールには、{name}のような関数やプロパティはありません',
    reason='{name}の打ち間違いでは？',
    solution='dir({module})で利用できる名前を調べてみよう.',
    inspect=inspect_callee,
    error_ext='{module}モジュールには、{name}のような{pyname}はありません',
    test=test_ModuleNoAttribute,
)


def test_ImportError():
    try:
        from math import sins
    except:
        kogi_check_error()


KOGI_ERR2(
    pattern='ImportError: cannot import name \'(.*?)\' from \'(.*?)\'',
    keys='name,module',
    error='モジュール{module}には{name}が見つかりません. ',
    reason='{name}の打ち間違いか、モジュール{module}の勘違い．',
    solution='dir({module})で利用できる名前を調べてみよう.',
    test=test_ImportError,
)


def test_UnsupportedOperand():
    try:
        print(1+'2')
    except:
        kogi_check_error()


KOGI_ERR2(
    pattern='unsupported operand type\(s\) for (.*?): \'(.*?)\' and \'(.*?)\'',
    keys='name,type,type2',
    error='{type}と{type2}の間で計算しようとしたけど、そこでは演算子{name}は使えません.',
    solution='{type}か{type2}のどちらかの型に変換するといいかも',
    inspect=inspect_infix,
    error_ext='{left}と{right}の間で計算しようとしたけど、型がそれぞれ{type}と{type2}で異なり、演算子{name}は使えません.',
    test=test_UnsupportedOperand,
)

KOGI_ERR2(
    pattern='\'(.*?)\' not supported between instances of \'(.*?)\' and \'(.*?)\'',
    keys='name,type,type2',
    error='{type}と{type2}の間で計算しようとしたけど、そこでは演算子{name}は使えません.',
    solution='{type}か{type2}のどちらかの型に変換するといいかも',
    inspect=inspect_infix,
    error_ext='{left}と{right}の間で計算しようとしたけど、型がそれぞれ{type}と{type2}で異なり、演算子{name}は使えません.',
)


def test_NotCallable():
    try:
        a = 1
        a()
    except:
        kogi_check_error()


KOGI_ERR2(
    pattern='\'(.*?)\' object is not callable',
    keys='type',
    error='{type}は、関数ではありません. ',
    reason='変数名に{type}の値を間違って代入してしまうと発生することがあります.',
    solution='環境を初期化してみてください',
    inspect=inspect_funcname,
    error_ext='{name}は、{type}型であり、関数ではありません. ',
    reason_ext='{name}に{type}の値を代入してしまうと、発生します.',
    test=test_NotCallable,
)


def test_NotIterable():
    try:
        list(1)
    except:
        kogi_check_error()


KOGI_ERR({
    'pattern': '\'(.*?)\' object is not iterable',
    'keys': 'type',
    'error': '{type}は、イテラブルではありません. つまり、リストに変換できませんし、for文などで繰り返し処理もできません.',
    'test': test_NotIterable,
})


def test_NotSubscriptable():
    try:
        a = 1
        a[0]
    except:
        kogi_check_error()


KOGI_ERR({
    'pattern': '\'(.*?)\' object is not subscriptable',
    'keys': 'type',
    'error': '{type}は、データ列でもマッピングでもありません.',
    'reason': 'たぶん、x[y]のように操作したけど、xは[]で値を取れません。',
    'test': test_NotSubscriptable,
})


def test_MustBeNoneOr():
    try:
        print(end=1)
    except:
        kogi_check_error()


KOGI_ERR({
    'pattern': '(\\w+?) must be None or a (\\w+?), not (\\w+)',
    'keys': 'name,type,type2',
    'error': '{name}は、Noneか{type}の値です. {type2}の値は使えません.',
    'test': test_MustBeNoneOr,
})


def test_InvalidKeyword():
    try:
        print(color='red')
    except:
        kogi_check_error()


KOGI_ERR2(
    pattern='\'(.*?)\' is an invalid keyword argument for (.*?)\\(\\)',
    keys='name,funcname',
    error='{name}は、{funcname}の引数として使えるキーワードではありません.',
    solution='{funcname}のリファレンスを読んで確認してください.',
    test=test_InvalidKeyword,
)

KOGI_ERR2(
    pattern='(.*?)\\(\\) got an unexpected keyword argument \'(.*?)\'',
    keys='funcname,name',
    error='{name}は、{funcname}の引数として使えるキーワードではありません.',
    solution='{funcname}のリファレンスを読んで確認してください.',
)


KOGI_ERR2(
    pattern='invalid syntax',
    keys='',
    error='構文エラーです',
    reason='Pythonは構文規則通りに書かなければなりません',
)

KOGI_ERR2(
    pattern='expected an indented block',
    keys='',
    error='インデントが足りません',
)

KOGI_ERR2(
    pattern='unexpected indent',
    keys='',
    error='インデントが余分です',
)

KOGI_ERR2(
    pattern='unindent does not match any outer indentation level',
    keys='',
    error='インデントの深さが変で、どのブロックに属すのかわかりません',
    solution='インデントの深さを揃えます',
)

KOGI_ERR2(
    pattern='invalid character in identifier',
    keys='',
    error='半角文字を使うべきところで、全角文字が混ざって使われています',
    solution='日本語入力をオフにして、打ち直そう',
    hint='先生やTAさんに質問する案件ではありません.',
)

KOGI_ERR2(
    pattern='unexpected EOF while parsing',
    keys='',
    error='コードが途中までしか書かれていません. ',
    reason='たぶん、括弧やクオートの閉じ忘れの可能性が高いです.',
    solution='エラーの発生した行の前後を含めて、構文エラーを探して！',
    hint='友達にチェックしてもらうとエラーが見つかるかも',
)


def test_InvalidLiteral():
    try:
        int('a')
    except:
        kogi_check_error()


KOGI_ERR({
    'pattern': 'invalid literal for int\(\) with base (.*?): (\'.*?\')',
    'keys': 'base,value',
    'error': '文字列{value}は、整数に変換できる文字列ではありません',
    'test': test_InvalidLiteral,
})


def test_NotConvert():
    try:
        float('a')
    except:
        kogi_check_error()


KOGI_ERR({
    'pattern': 'could not convert (\w+?) to (\w+?): \'(.*?)\'',
    'keys': 'type,type2,value',
    'error': '{type}の{value}は、{type2}に変換することはできません.',
    'test': test_NotConvert,
})


def test_OutOfRange():
    try:
        s = "ABC"
        s[3]
    except:
        kogi_check_error()


KOGI_ERR2(
    pattern='(\\w+) index out of range',
    keys='type',
    error='インデックスが{type}の大きさを超えています. ',
    inspect=inspect_index,
    error_ext='インデックス{index}が{callee}の大きさを超えています. ',
    solution_ext='インデックス{index}がlen({callee})未満に収まるように条件を加えてください. ',
    test=test_OutOfRange,
)


def test_KeyError():
    try:
        s = {}
        s["A"]
    except:
        kogi_check_error()


KOGI_ERR2(
    pattern='KeyError: (.+?)$',
    keys='key',
    error='キー{key}が見つかりません',
    inspect=inspect_index,
    error_ext='{callee}にはキー{key}がありません. ',
    reason_ext='{index}が未定義なキー{key}として与えられています. ',
    solution_ext='{callee}のキーを確認しましょう. ',
    test=test_KeyError,
)


def test_MustBeInteger():
    try:
        s = "ABC"
        s['3']
    except:
        kogi_check_error()


KOGI_ERR2(
    pattern='(\\w+) indices must be integers',
    keys='type',
    error='{type}のインデックスは、整数でなければなりません.',
    inspect=inspect_index,
    error_ext='{callee}のインデックスは、整数でなければなりません. ',
    reason_ext='{index}が整数でありません. ',
    solution_ext='{index}を整数に変換してみては？ ',
    test=test_MustBeInteger,
)


def test_DividedByZero():
    try:
        print(1/0)
    except:
        kogi_check_error()


KOGI_ERR2(
    pattern='division by zero',
    keys='',
    error='ゼロで割り算しました. ',
    reason='ゼロで割ることはできません. ',
    solution='分母の値を確認してみましょう.',
    test=test_DividedByZero
)


def test_FileNotFoundError():
    try:
        open('file')
    except:
        kogi_check_error()


KOGI_ERR2(
    pattern='FileNotFoundError: \\[Errno (\\d+)\\] No such file or directory: (\'.*\')',
    keys='errno,file',
    error='{file}が見つかりません. ',
    reason='ファイル名やファイルパスが間違っている. ',
    solution='{file}が存在するか確認しよう.',
    test=test_FileNotFoundError,
)


def test_DEFINED_ERRORS():
    for defined in DEFINED_ERRORS:
        if 'test' in defined:
            defined['test']()


if __name__ == '__main__':
    test_DEFINED_ERRORS()



from IPython.core.interactiveshell import InteractiveShell
from functools import wraps
import sys
#import traceback
import linecache
import json

from .exception_dialog import exception_dialog
from .logger import kogi_print, log

RUN_CELL = InteractiveShell.run_cell
SHOW_TRACEBACK = InteractiveShell.showtraceback
SHOW_SYNTAXERROR = InteractiveShell.showsyntaxerror


def stack_traceback(etype, emsg, tb):
    stacks = []
    while tb:
        filename = tb.tb_frame.f_code.co_filename
        if 'lib' not in filename:
            name = tb.tb_frame.f_code.co_name
            lineno = tb.tb_lineno
            line = linecache.getline(filename, lineno, tb.tb_frame.f_globals)
            local_vars = tb.tb_frame.f_locals
            stacks.append(dict(filename=filename,
                               etype=f'{etype.__name__}',
                               emsg=emsg,
                               name=name, lineno=lineno, line=line,
                               vars=local_vars))
        tb = tb.tb_next
    return list(stacks[::-1])


class KogiError(Exception):
    def __init__(self, **kw):
        Exception.__init__(self, json.dumps(kw))


def dummy_kogi_fn(raw_cell, emsg, stacks):
    kogi_print(raw_cell)
    kogi_print(emsg)
    kogi_print(stacks)


KOGI_FN = exception_dialog


def exception_hook(raw_cell, emsg, stacks):
    traceback = []
    for stack in stacks:
        d = stack.copy()
        del d['vars']
        traceback.append(d)
    log(
        type='exception_hook',
        code=raw_cell, emsg=emsg,
        traceback=traceback
    )
    try:
        KOGI_FN(raw_cell, emsg, stacks)
    except Exception as e:
        kogi_print(e)


def change_run_cell(func):
    @wraps(func)
    def run_cell(*args, **kwargs):
        if len(args) > 1:
            ipyshell = args[0]
            raw_cell = args[1]
            ipyshell.run_cell_raw_cell = raw_cell
        value = func(*args, **kwargs)
        return value
    return run_cell


def change_showtraceback(func):
    @wraps(func)
    def showtraceback(*args, **kwargs):
        etype, evalue, tb = sys.exc_info()
        emsg = f"{etype.__name__}: {evalue}"
        if not emsg.startswith('KogiError'):
            value = func(*args, **kwargs)
        else:
            value = None
        ipyshell = args[0]
        raw_cell = ipyshell.run_cell_raw_cell
        stacks = stack_traceback(etype, emsg, tb)
        exception_hook(raw_cell, emsg, stacks)
        return value
    return showtraceback


def enable_kogi_hook(kogi_fn=exception_dialog):
    global KOGI_FN
    KOGI_FN = kogi_fn
    InteractiveShell.run_cell = change_run_cell(RUN_CELL)
    InteractiveShell.showtraceback = change_showtraceback(SHOW_TRACEBACK)
    InteractiveShell.showsyntaxerror = change_showtraceback(SHOW_SYNTAXERROR)


def disable_kogi_hook():
    InteractiveShell.run_cell = RUN_CELL
    InteractiveShell.showtraceback = SHOW_TRACEBACK
    InteractiveShell.showsyntaxerror = SHOW_SYNTAXERROR
